import matplotlib.pyplot as plt
import numpy as np
import time
import sys
import random


###### PARTIE 1 ######

def construction_mphf(set_kmer:set[str], n:int, gamma:float=100, nb_niveaux:int=100) -> list[list[int]]:

	"""
	Construit une fonction de hachage minimale parfaite (MPHF) pour un ensemble de k-mers.

	Parameters:
	set_kmer (set): Ensemble de k-mers.
	n (int): Taille de l'ensemble de k-mers.
	gamma (int): Facteur de réduction de la taille de la table de hachage. Default: 2.
	nb_niveaux (int): Nombre de niveaux de réduction de la taille de la table de hachage. Default: 3.

	Returns:
	mphf (list): Table de hachage minimale parfaite.

	Examples:
	>>> set_kmer = {str(i) for i in range(10)}
	>>> n = 10
	>>> mphf = construction_mphf(set_kmer, n)
	>>> len(mphf) == n
	True
	>>> all(0 <= mphf[i] < n for i in range(n))
	True
	>>> len(mphf) == n
	True
	"""
	# Initialisation
	set_kmer_courant = set_kmer.copy()
	tableaux = []
	collision = set()
	for _ in range(nb_niveaux):
		if len(set_kmer_courant) > 0:
			l = len(set_kmer_courant)
			tableau_principal = [-1] * (gamma * l)
			for kmer in set_kmer_courant:
				h = abs(hash(kmer))
				address = h % (gamma * l)
				if tableau_principal[address] != -1:
					collision.add(kmer)
				else :
					tableau_principal[address] = h

			tableaux.append(tableau_principal) # Addition du tableau du dernier niveau dans un tableau contenant tous les tableaux des précédents niveaux
			set_kmer_courant = collision.copy() # Création d'un set de kmer à partir des collisions pour créer tableau du niveau suivant
			collision = set() # Vidage du set de collision pour qu'on puisse le remplir à l'itération suivante

	# Construction de la MPHF
	mphf = [] 
	grand_tableau = []
	for tableau in tableaux:
		grand_tableau.extend(tableau) # Concaténation de tous les tableaux de niveaux


	max_rang = 0

	for kmer in set_kmer:
		h = abs(hash(kmer))
		if h in grand_tableau:
			index = grand_tableau.index(h)
			rang = index - grand_tableau[:index + 1].count(-1)
			mphf.append([h, rang])
			max_rang = max(max_rang, rang)

	for kmer in set_kmer_courant: #gestion des collisions: expliquer les 3 lignes du dessous
		max_rang += 1 # augmentation du rang maximal pour l'attribuer à l'une des dernières collisions
		h = abs(hash(kmer)) # hachage de l'un des kmers restants dans les dernières collisions
		mphf.append([h, max_rang]) # ajout de l'un des kmers restants dans la fonction de hachage avec son rang associé
	
	return mphf


def get_hash_mphf(mphf, kmer):
	"""
	Calcule le hash d'un k-mer à l'aide d'une fonction de hachage minimale parfaite (MPHF).

	Parameters:
	mphf (list): Table de hachage minimale parfaite.
	kmer (str): K-mer à hasher.

	Returns:	int: Hash du k-mer.

	Examples:
	>>> set_kmer = {str(i) for i in range(10)}
	>>> n = 10
	>>> mphf = construction_mphf(set_kmer, n)
	>>> kmer = "5"
	>>> hash_mphf = get_hash_mphf(mphf, kmer)
	>>> 0 <= hash_mphf < n
	True
	"""
	h = abs(hash(kmer))
	for h_mphf,rang in mphf:
		if h_mphf == h :
			return rang

def create_hash_table(set_kmer, n):
	"""
	Crée une table de hachage à partir d'un ensemble de k-mers et d'une mphf

	Parameters:
	set_kmer (set): Ensemble de k-mers.
	n (int): Taille de la table de hachage.

	Returns:
	list: Table de hachage créée à partir des k-mers
	mphf: la mphf

	Examples:
	>>> set_kmer = {'ATCG', 'TGCA', 'GCTA'}
	>>> n = 10
	>>> tableau = create_hash_table(set_kmer, n)
	>>> len(tableau) == n
	True
	>>> all(kmer in tableau for kmer in set_kmer)
	True
	"""
	mphf = construction_mphf(set_kmer, n)
	table = [-1 for i in range(n)]
	for h_kmer, rang in mphf:
		table[rang] = h_kmer
	return table, mphf




def generer_kmers(n, k):
	'''
	genere un set de n k-mers
	'''
	kmers = set()
	while len(kmers) < n:
		kmer = ''.join(random.choice('ATCG') for _ in range(k))
		kmers.add(kmer)
	return kmers


def compare_taille(n_max, fichier_sortie):
	n_values = []
	table_size = []
	dict_size = []
	k = 21
	
	for n in range(100, n_max, 1000):
		set_kmer = generer_kmers(n, k)
		tableau, mphf = create_hash_table(set_kmer,n)

		n_values.append(n)
		table_size.append(sys.getsizeof(tableau)+sys.getsizeof(mphf)) # Comme le tableau nécessite une MPHF pour être généré, il est nécessaire de prendre en compte la taille de celle-ci
		dict_size.append(sys.getsizeof(set_kmer))

	plt.plot(n_values, table_size, label='Table avec MPHF')
	plt.plot(n_values, dict_size, label='Dict')
	plt.xlabel('n')
	plt.xticks(n_values, [str(x) for x in n_values], rotation=45)
	plt.ylabel('Taille (octets)')
	plt.title('Évolution de la taille de la table de hachage avec MPHF et du dict')
	plt.legend()
	plt.savefig(fichier_sortie)
	plt.close()

# Le dictionnaire a une gestion de l'espace en "escaliers", car lorsqu'il est plein il double son espace alloué pour réserver un grand nombre d'espcaes vides qui vont être remplis au fur et à mesure. Puis, lorsque cet espace vide atteint un certain seuil de remplissage, la taille double de nouveau.
# La table avec MPHF a une gestion de l'espace linéaire car elle ne réserve pas d'espace vide. En effet, la table est créée sur mesure pour avoir exactement l'espace requis pour stocker tous les élements.
compare_taille(10000,"mphf.png")
